import express from "express";
import multer from "multer";

import { initDb } from './db.js'
import { hash, compare } from 'bcrypt'
import { randomBytes } from "crypto";
import { copyFile, readFile } from "fs/promises";

const uploadStore = multer.diskStorage({
    destination: function (req, file, cb) {
        cb(null, './db/tmp')
    },
    filename: function (req, file, cb) {
        const uniqueSuffix = Date.now() + '-' + Math.round(Math.random() * 1E9)
        cb(null, file.fieldname + '-' + uniqueSuffix)
    }
});
const upload = multer({ storage: uploadStore });
const app = express()
const port = process.argv[2];

let db;
(async function () {
    db = await initDb();
})()

app.use(function (req, res, next) {
    res.setHeader("Access-Control-Allow-Origin", "*");
    res.setHeader('Access-Control-Allow-Methods', '*');
    res.setHeader("Access-Control-Allow-Headers", "*");
    next();
});

app.post('/register', upload.none(), async function (req, res) {
    const { user, pass, pass2 } = req.body;

    if (pass != pass2) {
        res.send({ status: 'Passwords do not match.' });
        return;
    }
    if (user.length < 1 || user.length > 64) {
        res.send({ status: 'Username too long.' })
        return;
    }
    if (/[^A-Za-z0-9\-\_]/g.test(user)) {
        res.send({ status: 'Invalid characters in username.' })
        return;
    }
    let accs = await db.all('SELECT username FROM auth WHERE UPPER(username) LIKE UPPER(?)', [
        user
    ]);

    if (accs && accs.length > 0) {
        res.send({ status: 'Account already exists' });
        return;
    }

    let hashed = await hash(pass, 10);

    await db.run('INSERT INTO auth (username, password) VALUES (?, ?)', [
        user,
        hashed
    ]);

    res.send({ redir: '/' });
});

app.post('/login', upload.none(), async function (req, res) {
    const { user, pass } = req.body;

    let accs = await db.all('SELECT * FROM auth WHERE UPPER(username) LIKE UPPER(?)', [
        user
    ]);

    if (!accs || accs.length < 1) {
        res.send({ status: 'Account does not exist.' });
        return;
    }

    let isHash = await compare(pass, accs[0].password);

    if (!isHash) {
        res.send({ status: 'Wrong account.' });
        return;
    }

    let token = randomBytes(256).toString('hex');

    await db.run('INSERT INTO token (username, token) VALUES (?, ?)', [
        user,
        token
    ])

    res.send({ redir: '/', token });
})

app.get('/token', async function (req, res) {
    let tokenIn = req.query.token;

    res.send(await getUser(tokenIn));
    return;
});

app.get('/user', async function (req, res) {
    let { user } = req.query;

    let following = await db.all('SELECT * FROM follow WHERE actor = ?', [
        user
    ]);

    let followers = await db.all('SELECT * FROM follow WHERE target = ?', [
        user
    ]);

    let userDat = (await db.all('SELECT * FROM user WHERE username = ?', [
        user
    ]))[0] || {};

    res.send({ id: user, following, followers, user: userDat });
});

app.post('/follow', upload.none(), async function (req, res) {
    let { user, tokenIn } = req.body;

    let { data } = await getUser(tokenIn);
    if (!data) {
        res.status(404);
        return;
    }

    let isFollow = await db.all('SELECT * FROM follow WHERE target = ? AND actor = ?', [
        user,
        data
    ]);

    if (isFollow && isFollow.length > 0) {
        await db.run('DELETE FROM follow WHERE target = ? AND actor = ?', [
            user,
            data
        ]);
    } else {
        await db.run('INSERT INTO follow (target, actor) VALUES (?,?)', [
            user,
            data
        ]);
    }
    await db.run('DELETE FROM follow WHERE actor IS NULL or actor NOT IN (SELECT username FROM auth)');
    await db.run('DELETE FROM follow WHERE target IS NULL or target NOT IN (SELECT username FROM auth)');
   
    res.send({});
});

app.post('/upload/studio', upload.none(), async function (req, res) {
    let { title, desc, admins, users, tokenIn } = req.body;

    let { data } = await getUser(tokenIn);
    if (!data) {
        res.status(404);
        return;
    }

    admins = admins.split(/\s+/g).filter(x => x != '');
    users = users.split(/\s+/g).filter(x => x != '');

    let isAdmin = true;

    let id = title.replaceAll(/[^0-9A-Za-z_-]/g, '_');
    let date = +(new Date);

    let baseStudio = (await db.all('SELECT * FROM studio WHERE id = ?', [
        id
    ]))[0];

    if (baseStudio) {
        isAdmin = !!(await db.all('SELECT * FROM studioUser WHERE username = ? AND studio = ? AND role = "admin"', [
            data,
            id
        ]))[0];
    }

    if (!isAdmin) {
        res.send({ redir: '/studios/' + id });
        return;
    }

    if (!baseStudio) {
        if (admins.length > 0 || users.length > 0) {
            await db.all('INSERT INTO studio (title, id, date, desc) VALUES (?, ?, ?, ?)', [
                title,
                id,
                date,
                desc
            ]);
        } else {
            return;
        }
    } else {
        await db.all('UPDATE studio SET desc = ? WHERE id = ?', [
            desc,
            id
        ]);
    }

    if (isAdmin && (admins.length > 0 || users.length > 0)) {
        await db.all('DELETE FROM studioUser WHERE studio = ?', [
            id
        ]);
        for (let admin of admins) {
            await db.all('DELETE FROM studioUser WHERE username = ?', [
                admin
            ]);
            await db.all('INSERT INTO studioUser (username, studio, role) VALUES (?, ?, ?)', [
                admin,
                id,
                'admin'
            ]);
        }
        for (let admin of users) {
            await db.all('DELETE FROM studioUser WHERE username = ?', [
                admin
            ]);
            await db.all('INSERT INTO studioUser (username, studio, role) VALUES (?, ?, ?)', [
                admin,
                id,
                'user'
            ]);
        }

    }
    
    await db.run('DELETE FROM studioUser WHERE username IS NULL or username NOT IN (SELECT username FROM auth)'); 
   
    res.send({ redir: '/studios/' + id });
})

app.post('/upload/project', upload.any(), async function (req, res) {
    let { title, unlisted, root, tokenIn } = req.body;
    let { path } = req.files[0];

    let { data } = await getUser(tokenIn);

    if (!data) {
        res.status(404);
        return;
    }

    root = parseInt(root) + '';
    
    let id = (await db.all('SELECT * from project')).length + 1;
    let date = +(new Date);

    let baseProject = (await db.all('SELECT * from project WHERE id = ?', [
        root
    ]))[0];

    if (baseProject && baseProject.username == data) {
        await db.run('UPDATE project SET title = ?, status = ? WHERE id = ?', [
            title,
            unlisted == 'on' ? 'unlisted' : null,
            baseProject.id
        ]);

        id = baseProject.id;

    } else {
        await db.run('INSERT INTO project (username, title, id, date, root, parent,status) VALUES (?,?,?,?,?,?,?)', [
            data,
            title,
            id,
            date,
            (baseProject) ? baseProject.root : id,
            (baseProject) ? root : null,
            unlisted == 'on' ? 'unlisted' : null
        ]);
    }

    await copyFile(path, `${process.cwd()}/db/projects/${id}`);

    if (req.files[1] && req.files[1].path)
        await copyFile(req.files[1].path, `${process.cwd()}/db/thumbs/${id}.png`);

    res.send({ redir: '/embed/' + id });
})

app.post('/upload/asset', upload.array('file'), async function (req, res) {
    let { files } = req;
    let { tokenIn } = req.body;

    let { data } = await getUser(tokenIn);
    if (!data) {
        res.status(404);
        return;
    }

    let filesR = [];

    for (let fileD of files) {
        let { path, filename } = fileD;

        filesR.push(filename);

        await copyFile(path, `${process.cwd()}/db/assets/${filename}`);
        await db.run('INSERT INTO asset (id, username) VALUES (?, ?)', [
            filename,
            data
        ]);
    }
    res.send({ redir: '/assets/', filesR });
});

app.post('/comment', upload.none(), async function (req, res) {
    let { type, id, content, tokenIn } = req.body;
    let { data } = await getUser(tokenIn);
    if (!data) {
        res.status(404);
        return;
    }
    content = content || '';
    let date = +(new Date);

    if (content.length < 1) return;

    let allComments = (await db.all('SELECT * FROM comment')).length;

    await db.run('INSERT INTO comment (username, targetType, targetId, date, content, id) VALUES (?,?,?,?,?,?)', [
        data,
        type,
        id,
        date,
        content,
        allComments + 1
    ]);

    res.send({ redir: `/${type}/${id}` });
})

app.get('/assets', upload.none(), async function (req, res) {
    let {  tokenIn } = req.query
    let { data } = await getUser(tokenIn);
    if (!data) {
        res.status(404);
        return;
    }

    let query = await db.all('SELECT * FROM asset WHERE username = ?', [
        data
    ]);

    res.send({ data: query });
})


app.post('/upload/bio', upload.none(), async function (req, res) {
    let { tokenIn, bio } = req.body;
    let { data } = await getUser(tokenIn);
    if (!data) {
        res.status(404);
        return;
    }

    let exists = (await db.all('SELECT * FROM user WHERE username = ? ', [
        data
    ]))[0];

    if (exists) {
        await db.run('UPDATE user SET bio = ? WHERE username = ?', [
            bio, data
        ]);
    } else {
        await db.run('INSERT INTO user (username,bio) VALUES (?,?)', [
            data, bio
        ]);
    }

    res.send({ redir: '/users/' + data });
});

app.post('/studio/add', upload.none(), async function (req, res) {
    let { tokenIn, studio, root } = req.body;
    let { data } = await getUser(tokenIn);

    if (!data) {
        res.status(404);
        return;
    }

    let isUser = (await db.all('SELECT * FROM studioUser WHERE username = ? AND studio = ?', [
        data,
        studio
    ]))[0];

    if (!isUser) return;

    await db.run('INSERT INTO studioProject (project, studio, role) VALUES (?,?,?)', [
        root,
        studio,
        'default'
    ])

    res.send({ redir: '/studios/' + studio });
});

app.get('/studio/', async function (req, res) {
    let { studio } = req.query;
    let dat = (await db.all('SELECT * FROM studioUser WHERE studio = ?', [
        studio
    ]));

    let name = (await db.all('SELECT * FROM studio WHERE id = ?', [
        studio
    ]))[0] || {};

    res.send({ id: studio, name: name.title, users: dat, bio: name.desc });
});

app.get('/list/studio', async function (req, res) {
    let { page } = req.query;

    let query = await db.all('SELECT * FROM studio ORDER BY date DESC LIMIT ?, ?', [
        page * 5,
        5
    ]);

    for (let i in query) {
        query[i].username = '';
    }

    res.send(query);
});

app.get('/raw', async function (req, res) {
    const legalSub = {
        'projects': '',
        'pfp': '.png',
        'assets': '',
        'thumbs': '.png'
    };
    let { sub, id } = req.query;
    id = id.replace(/[^A-Za-z0-9\-_]/g, '');

    if (Object.keys(legalSub).indexOf(sub) == -1) sub = 'projects';

    let path = `${process.cwd()}/db/${sub}/${id}${legalSub[sub]}`;
    let fe;
    try { 
        fe = await readFile(path);
    } catch (err) {
        return;
    }
    if ((fe.toString()).includes('<svg')) res.header('Content-Type', 'image/svg+xml');

    res.sendFile(path);
});

app.get('/embed', async function (req, res) {
    let { id } = req.query;

	let query = await db.all('SELECT * FROM project WHERE id = ?',[
		id
	]);

	res.send({id, data: query[0]});
})

app.get('/list/projects', async function (req, res) {
    let { param, page, type } = req.query;

    const types = {
        "studio": {
            'query': 'SELECT * FROM project WHERE id IN (SELECT project FROM studioProject WHERE studio = ?) AND date IS NOT "none" AND status IS NOT "unlisted" ORDER BY date DESC LIMIT ?, ?',
            'expanded': true
        },
        "user": {
            'query': 'SELECT * FROM project WHERE username = ? AND date IS NOT "none" AND status IS NOT "unlisted" ORDER BY date DESC LIMIT ?, ?',
            'expanded': true
        },
        "home": {
            'query': 'SELECT * FROM project WHERE date IS NOT "none" AND status IS NOT "unlisted" ORDER BY date DESC LIMIT ?, ?'
        },
        "feed": {
            'query': 'SELECT * FROM project WHERE username IN (SELECT target FROM follow WHERE actor = ?) AND date IS NOT "none" AND status IS NOT "unlisted" ORDER BY date DESC LIMIT ?, ?',
            'expanded': true
        }
    };

    let dat;

    if (types[type].expanded) {
        dat = [
            param,
            page * 5,
            5
        ]
    } else {
        dat = [
            page * 5,
            5
        ]
    }

    let query = await db.all(types[type].query, dat);

    res.send(query);
});

app.get('/list/comments', async function (req, res) {
    let { type, id, page } = req.query;

    let query = await db.all('SELECT * FROM comment WHERE targetType = ? AND targetId = ? ORDER BY date DESC LIMIT ?, ?', [
        type,
        id,
        page * 25,
        25
    ]);

    res.send(query);
})

async function getUser(tokenIn) {

    let acc = await db.all('SELECT username from token WHERE token = ?', [
        tokenIn
    ]);

    if (!acc || acc.length < 1) {
        return {};
    }

    let isBan = await db.all('SELECT * from ban WHERE username = ?', [
        acc[0].username
    ]);

    if (isBan && isBan.length > 0 && isBan[0].time > (+new Date)) {
        return { data: false, ban: isBan[0] };
    }

    return { data: acc[0].username, token: tokenIn };
}

app.listen(port, () => { })
